#!/usr/bin/env python
# Copyright (C) 2019 Emanuel Goncalves

import numpy as np
import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt
import matplotlib.colors as colors
import matplotlib.patches as mpatches
from natsort import natsorted
from crispy.DataImporter import Sample
from scipy.stats import gaussian_kde
from scipy.interpolate import interpn


class CrispyPlot:
    # PLOTING PROPS
    SNS_RC = {
        "axes.linewidth": 0.3,
        "xtick.major.width": 0.3,
        "ytick.major.width": 0.3,
        "xtick.major.size": 2.5,
        "ytick.major.size": 2.5,
        "xtick.minor.width": 0.3,
        "ytick.minor.width": 0.3,
        "xtick.minor.size": 1.5,
        "ytick.minor.size": 1.5,
        "xtick.direction": "out",
        "ytick.direction": "out",
    }

    # PALETTES
    PAL_DBGD = {0: "#656565", 1: "#F2C500", 2: "#E1E1E1", 3: "#0570b0"}
    PAL_SET1 = sns.color_palette("Set1", n_colors=9).as_hex()
    PAL_SET2 = sns.color_palette("Set2", n_colors=8).as_hex()
    PAL_DTRACE = {
        0: "#E1E1E1",
        1: PAL_SET2[1],
        2: "#656565",
        3: "#2b8cbe",
        4: "#de2d26",
    }

    PAL_YES_NO = {"No": "#E1E1E1", "Yes": PAL_SET2[1]}

    PAL_TISSUE = {
        "Lung": "#c50092",
        "Haematopoietic and Lymphoid": "#00ca5b",
        "Large Intestine": "#c100ce",
        "Central Nervous System": "#c3ce44",
        "Skin": "#d869ff",
        "Breast": "#496b00",
        "Head and Neck": "#ff5ff2",
        "Bone": "#004215",
        "Esophagus": "#fa006f",
        "Ovary": "#28d8fb",
        "Peripheral Nervous System": "#ff2c37",
        "Kidney": "#5ea5ff",
        "Stomach": "#fcbb3b",
        "Bladder": "#3e035f",
        "Pancreas": "#f6bc62",
        "Liver": "#df93ff",
        "Thyroid": "#ae7400",
        "Soft Tissue": "#f5aefa",
        "Cervix": "#008464",
        "Endometrium": "#980018",
        "Biliary Tract": "#535d3c",
        "Prostate": "#ff80b6",
        "Uterus": "#482800",
        "Vulva": "#ff7580",
        "Placenta": "#994800",
        "Testis": "#875960",
        "Small Intestine": "#fb8072",
        "Adrenal Gland": "#ff8155",
        "Eye": "#fa1243",
        "Other": "#000000",
    }

    PAL_TISSUE_2 = dict(
        zip(
            *(
                natsorted(list(PAL_TISSUE)),
                sns.color_palette("tab20c").as_hex()
                + sns.color_palette("tab20b").as_hex(),
            )
        )
    )

    PAL_CANCER_TYPE = dict(
        zip(
            *(
                natsorted(
                    list(Sample().samplesheet["cancer_type"].value_counts().index)
                ),
                sns.color_palette("tab20c").as_hex()
                + sns.color_palette("tab20b").as_hex()
                + sns.light_palette(PAL_SET1[1], n_colors=5).as_hex()[:-1],
            )
        )
    )
    PAL_CANCER_TYPE["Pancancer"] = PAL_SET1[5]

    PAL_MODEL_TYPE = {**PAL_CANCER_TYPE, **PAL_TISSUE_2}

    PAL_GROWTH_CONDITIONS = {
        "Adherent": "#fb8072",
        "Semi-Adherent": "#80b1d3",
        "Suspension": "#fdb462",
        "Unknown": "#d9d9d9",
    }

    PAL_MSS = {"MSS": "#d9d9d9", "MSI": "#fb8072"}

    SV_PALETTE = {
        "tandem-duplication": "#377eb8",
        "deletion": "#e41a1c",
        "translocation": "#984ea3",
        "inversion": "#4daf4a",
        "inversion_h_h": "#4daf4a",
        "inversion_t_t": "#ff7f00",
    }

    PPI_PAL = {
        "T": "#fc8d62",
        "1": "#656565",
        "2": "#7c7c7c",
        "3": "#949494",
        "4": "#ababab",
        "5+": "#c3c3c3",
        "-": "#2b8cbe",
        "X": "#2ca02c",
    }

    PPI_ORDER = ["T", "1", "2", "3", "4", "5+", "-"]

    GENESETS = ["essential", "nonessential", "nontargeting"]
    GENESETS_PAL = dict(
        essential="#e6550d", nonessential="#3182bd", nontargeting="#31a354"
    )

    # BOXPLOT PROPOS
    BOXPROPS = dict(linewidth=1.0)
    WHISKERPROPS = dict(linewidth=1.0)
    MEDIANPROPS = dict(linestyle="-", linewidth=1.0, color="red")
    FLIERPROPS = dict(
        marker="o",
        markerfacecolor="black",
        markersize=2.0,
        linestyle="none",
        markeredgecolor="none",
        alpha=0.6,
    )

    # CORRELATION PLOT PROPS
    ANNOT_KWS = dict(stat="R")
    MARGINAL_KWS = dict(kde=False, hist_kws={"linewidth": 0})

    LINE_KWS = dict(lw=1.0, color=PAL_DBGD[1], alpha=1.0)
    SCATTER_KWS = dict(edgecolor="w", lw=0.3, s=10, alpha=0.6, color=PAL_DBGD[0])
    JOINT_KWS = dict(lowess=True, scatter_kws=SCATTER_KWS, line_kws=LINE_KWS)

    @staticmethod
    def get_palette_continuous(n_colors, color=PAL_DBGD[0]):
        pal = sns.light_palette(color, n_colors=n_colors + 2).as_hex()[2:]
        return pal

    @staticmethod
    def density_interpolate(xx, yy, dtype="gaussian"):
        if dtype == "gaussian":
            xy = np.vstack([xx, yy])
            zz = gaussian_kde(xy)(xy)

        else:
            data, x_e, y_e = np.histogram2d(xx, yy, bins=20)

            zz = interpn(
                (0.5 * (x_e[1:] + x_e[:-1]), 0.5 * (y_e[1:] + y_e[:-1])),
                data,
                np.vstack([xx, yy]).T,
                method="splinef2d",
                bounds_error=False,
            )

        return zz

    @classmethod
    def get_palettes(cls, samples, samplesheet):
        samples = set(samples).intersection(samplesheet.index)

        pal_tissue = {
            s: cls.PAL_TISSUE_2[samplesheet.loc[s, "tissue"]] for s in samples
        }

        pal_growth = {
            s: cls.PAL_GROWTH_CONDITIONS[samplesheet.loc[s, "growth_properties"]]
            for s in samples
        }

        palettes = pd.DataFrame(dict(tissue=pal_tissue, media=pal_growth))

        return palettes

    @classmethod
    def triu_plot(cls, x, y, color, label, **kwargs):
        df = pd.DataFrame(dict(x=x, y=y)).dropna()
        ax = plt.gca()
        ax.hexbin(
            df["x"], df["y"], cmap="Spectral_r", gridsize=30, mincnt=1, bins="log", lw=0
        )
        ax.grid(True, ls=":", lw=0.1, alpha=1.0, zorder=0)

    @classmethod
    def triu_scatter_plot(cls, x, y, color, label, **kwargs):
        df = pd.DataFrame(dict(x=x, y=y)).dropna()
        df["z"] = cls.density_interpolate(df["x"], df["y"], dtype="interpolate")
        df = df.sort_values("z")

        ax = plt.gca()

        ax.scatter(
            df["x"],
            df["y"],
            c=df["z"],
            marker="o",
            edgecolor="",
            s=5,
            alpha=0.8,
            cmap="Spectral_r",
        )

        ax.grid(True, ls=":", lw=0.1, alpha=1.0, zorder=0)

    @classmethod
    def diag_plot(cls, x, color, label, **kwargs):
        sns.distplot(
            x[~np.isnan(x)], label=label, color=CrispyPlot.PAL_DBGD[0], kde=False
        )

    @classmethod
    def attenuation_scatter(
        cls,
        x,
        y,
        plot_df,
        z="cluster",
        zorder=None,
        pal=None,
        figsize=(2.5, 2.5),
        plot_reg=True,
        ax_min=None,
        ax_max=None,
    ):
        if ax_min is None:
            ax_min = plot_df[[x, y]].min().min() * 1.1

        if ax_max is None:
            ax_max = plot_df[[x, y]].max().max() * 1.1

        if zorder is None:
            zorder = ["High", "Low"]

        if pal is None:
            pal = dict(High=cls.PAL_DTRACE[1], Low=cls.PAL_DTRACE[0])

        g = sns.jointplot(
            x,
            y,
            plot_df,
            "scatter",
            color=CrispyPlot.PAL_DTRACE[0],
            xlim=[ax_min, ax_max],
            ylim=[ax_min, ax_max],
            space=0,
            s=5,
            edgecolor="w",
            linewidth=0.0,
            marginal_kws={"hist": False, "rug": False},
            stat_func=None,
            alpha=0.1,
        )

        for n in zorder[::-1]:
            df = plot_df.query(f"{z} == '{n}'")
            g.x, g.y = df[x], df[y]

            g.plot_joint(
                sns.regplot,
                color=pal[n],
                fit_reg=False,
                scatter_kws={"s": 3, "alpha": 0.5, "linewidth": 0},
            )

            if plot_reg:
                g.plot_joint(
                    sns.kdeplot,
                    cmap=sns.light_palette(pal[n], as_cmap=True),
                    legend=False,
                    shade=False,
                    shade_lowest=False,
                    n_levels=9,
                    alpha=0.8,
                    lw=0.1,
                )

            g.plot_marginals(sns.kdeplot, color=pal[n], shade=True, legend=False)

        handles = [mpatches.Circle([0, 0], 0.25, facecolor=pal[s], label=s) for s in pal]
        g.ax_joint.legend(
            loc="upper left", handles=handles, title="Protein\nattenuation", frameon=False
        )

        g.ax_joint.grid(True, ls="-", lw=0.1, alpha=1.0, zorder=0)
        g.ax_joint.plot([ax_min, ax_max], [ax_min, ax_max], "k--", lw=0.3)

        plt.gcf().set_size_inches(figsize)

        return g


class MidpointNormalize(colors.Normalize):
    def __init__(self, vmin=None, vmax=None, midpoint=None, clip=False):
        self.midpoint = midpoint
        colors.Normalize.__init__(self, vmin, vmax, clip)

    def __call__(self, value, clip=None):
        x, y = [self.vmin, self.midpoint, self.vmax], [0, 0.5, 1]
        return np.ma.masked_array(np.interp(value, x, y))



#!/usr/bin/env python
# Copyright (C) 2019 Emanuel Goncalves

import numpy as np
import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt
from crispy.CrispyPlot import CrispyPlot


class LibraryRepresentaion:
    def __init__(self, counts, percentiles=None):
        self.counts = counts
        self.percentiles = [2.5, 97.5] if percentiles is None else percentiles

    def gini(self):
        gini_scores = {}

        for c in self.counts:
            values = self.counts[c].dropna()

            sorted_list = sorted(values)

            height, area = 0.0, 0.0

            for value in sorted_list:
                height += value
                area += height - value / 2.0

            fair_area = height * len(values) / 2.0

            gini_scores[c] = (fair_area - area) / fair_area

        return pd.Series(gini_scores)

    def percentile(self):
        percentile_ranges = {}

        for c in self.counts:
            values = self.counts[c]

            p_low, p_high = np.nanpercentile(np.log10(values), self.percentiles)

            percentile_range = p_high / p_low

            percentile_ranges[c] = percentile_range

        return pd.Series(percentile_ranges)

    def dropout_rate(self, threshold=0):
        dropout_rates = {}

        for c in self.counts:
            values = self.counts[c]
            dropout_rates[c] = sum(values <= threshold) / len(values)

        return pd.Series(dropout_rates)

    def boxplot(self, order=None, palette=None, ax=None):
        values = np.log10(self.counts)

        if ax is None:
            ax = plt.gca()

        sns.boxplot(
            data=values,
            order=order,
            orient="h",
            palette=palette,
            saturation=1,
            showcaps=False,
            boxprops=dict(linewidth=0.3),
            whiskerprops=dict(linewidth=0.3),
            flierprops=dict(
                marker="o",
                markerfacecolor="black",
                markersize=1.0,
                linestyle="none",
                markeredgecolor="none",
                alpha=0.6,
            ),
            notch=True,
            ax=ax,
        )

        ax.grid(True, ls="-", lw=0.1, alpha=1.0, zorder=0)

        ax.set_xlabel("sgRNAs counts (log10)")

        return ax

    def distplot(self, palette=None, ax=None):
        if ax is None:
            ax = plt.gca()

        for c in self.counts:
            values = self.counts[c]

            sns.distplot(
                values,
                hist=False,
                kde_kws=dict(cut=0),
                label=c,
                color=None if palette is None else palette[c],
                ax=ax,
            )

        ax.grid(True, ls="-", lw=0.1, alpha=1.0, zorder=0)

        ax.set_xlabel("sgRNA counts")

        plt.legend(frameon=False, prop={'size': 4})

        return ax

    def lorenz_curve(self, palette=None, ax=None):
        if ax is None:
            ax = plt.gca()

        gini_scores = self.gini()

        for c in self.counts:
            values = np.sort(self.counts[c].dropna())

            x_lorenz = np.cumsum(values) / np.sum(values)
            y_lorenz = np.arange(1, len(x_lorenz) + 1) / len(x_lorenz)

            ax.plot(
                y_lorenz,
                x_lorenz,
                color=CrispyPlot.PAL_DBGD[0] if palette is None else palette[c],
                label=f"{c}\n(Gini coefficient={gini_scores[c]:.2f})",
                zorder=2,
                alpha=.5,
            )

        ax.plot([0, 1], [0, 1], "k-", lw=0.3, zorder=1, label="Expected line of equality")

        ax.grid(True, ls="-", lw=0.1, alpha=1.0, zorder=0)

        ax.set_ylabel("Fraction of total reads")
        ax.set_xlabel("Fraction of total sgRNAs")

        plt.legend(frameon=False, prop={"size": 4}, loc="center left", bbox_to_anchor=(1, 0.5))

        return ax



#!/usr/bin/env python
# Copyright (C) 2019 Emanuel Goncalves

import ast
import numpy as np
import pandas as pd
from crispy import logger as LOG
from crispy.CRISPRData import Library


class GuideSelection:
    GREEN = dict(offtarget=[1, 0], jacks_thres=1, ruleset2_thres=0.4)
    AMBER = dict(offtarget=[1], jacks_thres=None, ruleset2_thres=None)
    RED = dict(offtarget=[3])

    def __init__(self, masterlib="MasterLib_v1.csv.gz"):
        self.masterlibfile = masterlib
        self.masterlib = self.import_master_library()

    def import_master_library(self):
        df = Library.load_library(self.masterlibfile, set_index=False)

        # Drop sgRNAs that do not align to GRCh38
        df = df.query("Assembly == 'Human (GRCh38)'")

        # Remove sgRNAs that match to multiple genes
        sg_count = df.groupby("WGE_Sequence")["Approved_Symbol"].agg(
            lambda v: len(set(v))
        )
        sg_count = sg_count[sg_count != 1]
        df = df[~df["WGE_Sequence"].isin(sg_count.index)]

        # Remove sgRNAs with no alignment to GRCh38
        df = df.dropna(subset=["Approved_Symbol", "Off_Target"])

        # Remove sgRNAs with U6 stop codon
        df = df[["TTTT" not in g for g in df["WGE_Sequence"]]]

        # Remove duplicates (sgRNAs shared across multiple libraries)
        df["Library"] = pd.Categorical(
            df["Library"], ["KosukeYusa", "Avana", "Brunello", "TKOv3"]
        )
        df = df.sort_values("Library")
        df = df.groupby("WGE_Sequence").first().reset_index()

        # Parse off target summaries
        df["Off_Target"] = df["Off_Target"].apply(ast.literal_eval).values

        # Calculate absolute distance of JACKS scores to 1 (similar to gene sgRNAs mean)
        df["JACKS_min"] = abs(df["JACKS"] - 1).values

        # Sort guides according to KS scores
        df = df.sort_values(
            ["KS", "JACKS_min", "RuleSet2"], ascending=[False, True, False]
        )

        return df

    def get_sgrnas(
        self,
        gene,
        offtarget,
        n_guides,
        library=None,
        sortby=None,
        ascending=None,
        query=None,
        dropna=None,
        sgrnas_exclude=None,
    ):
        # Select gene sgRNAs gene
        gguides = self.masterlib.query(f"Approved_Symbol == '{gene}'")

        # Subset to library
        if library is not None:
            gguides = gguides.query(f"Library == '{library}'")

        # List of sgRNAs to exclude
        if sgrnas_exclude is not None:
            gguides = gguides[~gguides["sgRNA_ID"].isin(sgrnas_exclude)]

        # Drop NaNs
        if dropna is not None:
            gguides = gguides.dropna(subset=dropna)

        # Query
        if query is not None:
            gguides = gguides.query(query)

        # Sort guides
        if (sortby is not None) and (ascending is not None):
            gguides = gguides.sort_values(sortby, ascending=ascending)

        # Off-target filter
        gguides = gguides.loc[
            [
                np.all([ot[i] <= v for i, v in enumerate(offtarget)])
                for ot in gguides["Off_Target"]
            ]
        ]

        # Pick top n sgRNAs
        gguides = gguides.head(n_guides)

        return gguides

    def select_sgrnas(
        self,
        gene,
        n_guides=5,
        jacks_thres=1,
        ruleset2_thres=0.4,
        offtarget=None,
        sgrnas_exclude=None,
    ):
        offtarget = [1, 0] if offtarget is None else offtarget

        # KosukeYusa v1.1 sgRNAs
        g_guides = self.get_sgrnas(
            gene=gene,
            library="KosukeYusa",
            offtarget=offtarget,
            n_guides=n_guides,
            query=f"JACKS_min < {jacks_thres}" if jacks_thres is not None else None,
            dropna=["KS"],
            sgrnas_exclude=sgrnas_exclude,
        )

        # Top-up with Avana guides
        if len(g_guides) < n_guides:
            g_guides_avana = self.get_sgrnas(
                gene=gene,
                library="Avana",
                offtarget=offtarget,
                n_guides=n_guides - g_guides.shape[0],
                query=f"JACKS_min < {jacks_thres}" if jacks_thres is not None else None,
                dropna=["KS"],
                sgrnas_exclude=sgrnas_exclude,
            )
            g_guides = pd.concat([g_guides, g_guides_avana], sort=False)

        # Top-up with Brunello guides
        if len(g_guides) < n_guides:
            g_guides_brunello = self.get_sgrnas(
                gene=gene,
                library="Brunello",
                offtarget=offtarget,
                n_guides=n_guides - g_guides.shape[0],
                query=f"RuleSet2 > {ruleset2_thres}" if ruleset2_thres is not None else None,
                sortby="RuleSet2",
                ascending=False,
                sgrnas_exclude=sgrnas_exclude,
            )
            g_guides = pd.concat([g_guides, g_guides_brunello], sort=False)

        # Top-up with TKOv3 guides
        if len(g_guides) < n_guides:
            g_guides_tkov3 = self.get_sgrnas(
                gene=gene,
                library="TKOv3",
                offtarget=offtarget,
                n_guides=n_guides - g_guides.shape[0],
                sgrnas_exclude=sgrnas_exclude,
            )
            g_guides = pd.concat([g_guides, g_guides_tkov3], sort=False)

        return g_guides

    def selection_rounds(self, gene, n_guides=5, do_amber_round=True, do_red_round=True, verbose=1):
        # - Stringest sgRNA selection round
        gguides = self.select_sgrnas(gene=gene, n_guides=n_guides, **self.GREEN)
        gguides = gguides.assign(Confidence="Green")

        # - Middle sgRNA selection round
        if do_amber_round and (len(gguides) < n_guides):
            gguides_amber = self.select_sgrnas(
                gene=gene,
                n_guides=n_guides - gguides.shape[0],
                sgrnas_exclude=set(gguides["sgRNA_ID"]),
                **self.AMBER,
            ).assign(Confidence="Amber")
            gguides = pd.concat([gguides, gguides_amber], sort=False)

            if verbose > 0:
                LOG.warn(f"{len(gguides_amber)} Amber guides selected for gene {gene}")

        # - Lose sgRNA selection round
        if do_red_round and (len(gguides) < n_guides):
            gguides_red = self.get_sgrnas(
                gene=gene,
                n_guides=n_guides - gguides.shape[0],
                sgrnas_exclude=set(gguides["sgRNA_ID"]),
                **self.RED,
            ).assign(Confidence="Red")
            gguides = pd.concat([gguides, gguides_red], sort=False)

            if verbose > 0:
                LOG.warn(f"{len(gguides_red)} Red guides selected for gene {gene}")

        return gguides



#!/usr/bin/env python
# Copyright (C) 2019 Emanuel Goncalves

import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()


version = {}
with open("crispy/__init__.py") as f:
    exec(f.read(), version)


with open("requirements.txt") as f:
    requirements = f.read().split("\n")


included_files = {
    "crispy": [
        "data/cytoBand.txt",

        "data/crispr_libs/Avana_v1.csv.gz",
        "data/crispr_libs/Brunello_v1.csv.gz",
        "data/crispr_libs/GeCKO_v2.csv.gz",
        "data/crispr_libs/Manjunath_Wu_v1.csv.gz",
        "data/crispr_libs/MasterLib_v1.csv.gz",
        "data/crispr_libs/MinLibCas9.csv.gz",
        "data/crispr_libs/Sabatini_Lander_v3.csv.gz",
        "data/crispr_libs/TKOv3.csv.gz",
        "data/crispr_libs/Yusa_v1.1.csv.gz",

        "data/crispr_manifests/project_score_exclude_samples.csv",
        "data/crispr_manifests/GeCKO2_Achilles_v3.3.8_dropped_guides.csv.gz",
        "data/crispr_manifests/Avana_DepMap19Q2_sample_map.csv.gz",
        "data/crispr_manifests/Avana_DepMap19Q2_dropped_guides.csv.gz",
        "data/crispr_manifests/Avana_DepMap19Q3_sample_map.csv.gz",
        "data/crispr_manifests/Avana_DepMap19Q3_dropped_guides.csv",
        "data/crispr_manifests/KM12_coverage_samplesheet.xlsx",

        "data/gene_sets/EssentialGenes.csv",
        "data/gene_sets/NonessentialGenes.csv",
        "data/gene_sets/pancan_core.csv",
        "data/gene_sets/curated_BAGEL_essential.csv",
        "data/gene_sets/curated_BAGEL_nonEssential.csv",
        "data/gene_sets/pan_core_essential_broad_depmap18Q3.txt",
        "data/gene_sets/pan_core_essential_broad_depmap18Q4.txt",
        "data/gene_sets/essential_sanger_depmap19.tsv",

        "data/copy_number/Summary_segmentation_data_994_lines_picnic.csv.gz",

        "data/images/example_gp_fit.png",
        "data/example_rawcounts.csv",
        "data/example_copynumber.csv",

        "data/images/logo.png",

        "data/model_list_20200204.csv",
        "data/GrowthRates_v1.3.0_20190222.csv",
        "data/growth_rates_rapid_screen_1536_v1.6.3_02Jun20.csv",
        "data/SIDMvsMedia.xlsx",
        "data/crispr/CRISPR_Institute_Origin_20191108.csv.gz",

        "data/pathways/c2.all.v7.1.symbols.gmt",
        "data/pathways/c2.cp.kegg.v7.1.symbols.gmt",
        "data/pathways/c2.cp.v7.1.symbols.gmt",
        "data/pathways/c5.all.v7.1.symbols.gmt",
        "data/pathways/c5.bp.v7.1.symbols.gmt",
        "data/pathways/c6.all.v7.1.symbols.gmt",
        "data/pathways/h.all.v7.1.symbols.gmt",
    ]
}


setuptools.setup(
    name="cy",
    version=version["__version__"],
    author="Emanuel Goncalves",
    author_email="eg14@sanger.ac.uk",
    long_description=long_description,
    description="Modelling CRISPR dropout data",
    long_description_content_type="text/markdown",
    url="https://github.com/EmanuelGoncalves/crispy",
    packages=["crispy", "test"],
    include_package_data=True,
    package_data=included_files,
    install_requires=requirements,
    classifiers=(
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
        "Topic :: Scientific/Engineering :: Information Analysis",
    ),
)



#!/usr/bin/env python
# Copyright (C) 2019 Emanuel Goncalves

import logging
import numpy as np
import pandas as pd
import pkg_resources
import seaborn as sns
import matplotlib.pyplot as plt
from crispy.CrispyPlot import CrispyPlot
from scipy import stats
from natsort import natsorted
from sklearn.metrics import roc_curve, roc_auc_score
from sklearn.model_selection import StratifiedShuffleSplit

LOG = logging.getLogger("Crispy")
DPATH = pkg_resources.resource_filename("crispy", "data/")


class GExp:
    GEXP_GENESETS_FILE = f"{DPATH}/GExp_genesets_20191126.csv"

    def __init__(self, genesets=None):
        if genesets is None:
            self.genesets = (
                pd.read_csv(self.GEXP_GENESETS_FILE)
                .groupby("gtype")["gene"]
                .agg(set)
                .to_dict()
            )

        else:
            self.genesets = genesets

        LOG.info(
            f"N Genes low/high: {len(self.genesets['low'])}/{len(self.genesets['high'])}"
        )

        self.palette = {"low": "#fc8d62", "high": "#2b8cbe", "all": "#656565"}

    def plot_histogram(self, x, figsize=(2.5, 1.5)):
        _, ax = plt.subplots(1, 1, figsize=figsize)
        for n, s in [
            ("low", x.reindex(self.genesets["low"])),
            ("high", x.reindex(self.genesets["high"])),
            ("all", x),
        ]:
            sns.distplot(
                s,
                hist=False,
                label=n,
                kde_kws={"cut": 0, "shade": True},
                color=self.palette[n],
                ax=ax,
            )
        ax.grid(True, ls="-", lw=0.1, alpha=1.0, zorder=0, axis="x")
        ax.legend(frameon=False, prop={"size": 5})
        ax.set_title(x.name)
        return ax

    @staticmethod
    def benchmark(y_true, y_score, max_fpr, ax=None):
        fpr, tpr, thres = roc_curve(y_true, y_score)

        auc = roc_auc_score(y_true, y_score)

        fpr_thres = min(thres[fpr <= max_fpr])

        if ax is not None:
            ax.plot(fpr, tpr, c=CrispyPlot.PAL_DBGD[0])

        return auc, fpr_thres

    def discretise(
        self,
        x,
        n_splits=100,
        train_size=0.7,
        decimal_places=2,
        max_fpr=0.05,
        genesets=None,
        verbose=0,
    ):
        genesets = self.genesets if genesets is None else genesets

        # Round x values
        x = x.round(decimal_places)

        # Gene-expression values range
        x_range = list(natsorted(set(x)))

        # x values labels
        x_labels = pd.Series([0] * len(x), index=x.index)
        x_labels.loc[x_labels.index.isin(genesets["low"])] = -1
        x_labels.loc[x_labels.index.isin(genesets["high"])] = 1

        # Verbose plots
        axs = plt.subplots(1, 2, figsize=(8.0, 4.0), dpi=600)[1] if verbose > 1 else [None, None]

        # Cross-validation
        x_lr, x_stats = [], []

        cv = StratifiedShuffleSplit(n_splits, train_size=train_size)
        for i, (idx_train, idx_test) in enumerate(cv.split(x.to_frame(), x_labels)):
            # Train and Test set
            x_train, x_test = x.iloc[idx_train], x.iloc[idx_test]

            # High and low gene-expression values
            x_train_low = x_train.reindex(genesets["low"]).dropna()
            x_train_high = x_train.reindex(genesets["high"]).dropna()

            if verbose > 0:
                LOG.info(
                    f"[Iteration {i+1}] N Genes (low/high): {len(x_train_low)}/{len(x_train_high)}"
                )

            # Density fitting
            x_train_low_k = stats.gaussian_kde(x_train_low)
            x_train_high_k = stats.gaussian_kde(x_train_high)

            # Estimate log-ratios for values range
            lr = pd.Series(
                np.log2(
                    x_train_high_k.evaluate(x_range) / x_train_low_k.evaluate(x_range)
                ),
                index=x_range,
            )

            # Assign x_test log-ratios
            x_test_lr = lr[x_test.values]
            x_test_lr.index = x_test.index
            x_lr.append(x_test_lr)

            # Benchmark
            for i, gset in enumerate(genesets):
                y_test_score = x_test_lr if gset == "high" else -x_test_lr
                y_test_true = y_test_score.index.isin(genesets[gset]).astype(int)

                y_test_auc, y_test_thres = self.benchmark(
                    y_test_true, y_test_score, max_fpr=max_fpr, ax=axs[i]
                )
                y_test_thres = y_test_thres if gset == "high" else -y_test_thres

                x_stats.append(dict(auc=y_test_auc, thres=y_test_thres, gtype=gset))

        # Verbose plots
        if verbose > 1:
            for i, gset in enumerate(genesets):
                axs[i].set_xlabel("False Positive Rate")
                axs[i].set_ylabel("True Positive Rate")
                axs[i].set_title(f"Geneset = {gset.capitalize()}")
                axs[i].grid(True, ls="-", lw=0.1, alpha=1.0, zorder=0)
                axs[i].plot((0, 1), (0, 1), "k-", lw=0.3, zorder=0)

            plt.subplots_adjust(hspace=0.05, wspace=0.05)
            plt.show()

        # Aggregate
        x_lr = pd.concat(x_lr, axis=1, sort=False)
        x_stats = pd.DataFrame(x_stats)

        # Returning data-frame
        res = pd.concat(
            [
                x_lr.mean(1).rename("lr_mean"),
                x_lr.std(1).rename("lr_std"),
                x_lr.count(1).rename("lr_count"),
                x_labels.rename("gtype"),
            ],
            axis=1,
            sort=False,
        ).sort_values("lr_mean")

        for gset, gset_df in x_stats.groupby("gtype"):
            res[f"{gset}_thres_std"] = gset_df["thres"].std()
            res[f"{gset}_thres_mean"] = gset_df["thres"].mean()

            if gset == "high":
                res[gset] = (res["lr_mean"] > gset_df["thres"].mean()).astype(int)

            else:
                res[gset] = (res["lr_mean"] < gset_df["thres"].mean()).astype(int)

        if (len(x) - len(x_lr)) != 0:
            LOG.warning(f"Genes {len(x)} / {len(res)} cross-validated")

        return res



#!/usr/bin/env python
# Copyright (C) 2019 Emanuel Goncalves

import numpy as np
import pandas as pd
import pkg_resources
from scipy.stats import gaussian_kde, rankdata, norm, spearmanr, pearsonr


class Utils(object):
    DPATH = pkg_resources.resource_filename("crispy", "data/")

    CHR_ORDER = [
        "1",
        "2",
        "3",
        "4",
        "5",
        "6",
        "7",
        "8",
        "9",
        "10",
        "11",
        "12",
        "13",
        "14",
        "15",
        "16",
        "17",
        "18",
        "19",
        "20",
        "21",
        "22",
        "X",
        "Y",
    ]

    CHR_SIZES_HG19 = {
        "chr1": 249_250_621,
        "chr2": 243_199_373,
        "chr3": 198_022_430,
        "chr4": 191_154_276,
        "chr5": 180_915_260,
        "chr6": 171_115_067,
        "chr7": 159_138_663,
        "chr8": 146_364_022,
        "chr9": 141_213_431,
        "chr10": 135_534_747,
        "chr11": 135_006_516,
        "chr12": 133_851_895,
        "chr13": 115_169_878,
        "chr14": 107_349_540,
        "chr15": 102_531_392,
        "chr16": 90_354_753,
        "chr17": 81_195_210,
        "chr18": 78_077_248,
        "chr19": 59_128_983,
        "chr20": 63_025_520,
        "chr21": 48_129_895,
        "chr22": 51_304_566,
        "chrX": 155_270_560,
        "chrY": 59_373_566,
    }

    BRASS_HEADERS = [
        "chr1",
        "start1",
        "end1",
        "chr2",
        "start2",
        "end2",
        "id/name",
        "brass_score",
        "strand1",
        "strand2",
        "sample",
        "svclass",
        "bkdist",
        "assembly_score",
        "readpair names",
        "readpair count",
        "bal_trans",
        "inv",
        "occL",
        "occH",
        "copynumber_flag",
        "range_blat",
        "Brass Notation",
        "non-template",
        "micro-homology",
        "assembled readnames",
        "assembled read count",
        "gene1",
        "gene_id1",
        "transcript_id1",
        "strand1",
        "end_phase1",
        "region1",
        "region_number1",
        "total_region_count1",
        "first/last1",
        "gene2",
        "gene_id2",
        "transcript_id2",
        "strand2",
        "phase2",
        "region2",
        "region_number2",
        "total_region_count2",
        "first/last2",
        "fusion_flag",
    ]

    ASCAT_HEADERS = ["chr", "start", "end", "copy_number"]

    @staticmethod
    def bin_bkdist(distance):
        """
        Discretise genomic distances

        :param distance: float

        :return:

        """
        if distance == -1:
            bin_distance = "diff. chr."

        elif distance == 0:
            bin_distance = "0"

        elif 1 < distance < 10e3:
            bin_distance = "1-10 kb"

        elif 10e3 < distance < 100e3:
            bin_distance = "10-100 kb"

        elif 100e3 < distance < 1e6:
            bin_distance = "0.1-1 Mb"

        elif 1e6 < distance < 10e6:
            bin_distance = "1-10 Mb"

        else:
            bin_distance = ">10 Mb"

        return bin_distance

    @staticmethod
    def svtype(strand1, strand2, svclass, unfold_inversions):
        if svclass == "translocation":
            svtype = "translocation"

        elif strand1 == "+" and strand2 == "+":
            svtype = "deletion"

        elif strand1 == "-" and strand2 == "-":
            svtype = "tandem-duplication"

        elif strand1 == "+" and strand2 == "-":
            svtype = "inversion_h_h" if unfold_inversions else "inversion"

        elif strand1 == "-" and strand2 == "+":
            svtype = "inversion_t_t" if unfold_inversions else "inversion"

        else:
            assert (
                False
            ), "SV class not recognised: strand1 == {}; strand2 == {}; svclass == {}".format(
                strand1, strand2, svclass
            )

        return svtype

    @staticmethod
    def bin_cnv(value, thresold):
        if not np.isfinite(value):
            return np.nan

        value = int(round(value, 0))
        value = f"{value}" if value < thresold else f"{thresold}+"

        return value

    @staticmethod
    def qnorm(x):
        y = rankdata(x)
        y = -norm.isf(y / (len(x) + 1))
        return y

    @classmethod
    def get_example_data(cls):
        raw_counts = pd.read_csv(
            "{}/{}".format(cls.DPATH, "example_rawcounts.csv"), index_col=0
        )
        copynumber = pd.read_csv("{}/{}".format(cls.DPATH, "example_copynumber.csv"))
        return raw_counts, copynumber

    @classmethod
    def get_essential_genes(
        cls, dfile="gene_sets/EssentialGenes.csv", return_series=True
    ):
        geneset = set(pd.read_csv(f"{cls.DPATH}/{dfile}", sep="\t")["gene"])

        if return_series:
            geneset = pd.Series(list(geneset)).rename("essential")

        return geneset

    @classmethod
    def get_non_essential_genes(
        cls, dfile="gene_sets/NonessentialGenes.csv", return_series=True
    ):
        geneset = set(pd.read_csv(f"{cls.DPATH}/{dfile}", sep="\t")["gene"])

        if return_series:
            geneset = pd.Series(list(geneset)).rename("non-essential")

        return geneset

    @classmethod
    def get_crispr_lib(cls, dfile="crispr_libs/Yusa_v1.1.csv.gz"):
        r_cols = dict(
            index="sgrna",
            CHRM="chr",
            STARTpos="start",
            ENDpos="end",
            GENES="gene",
            EXONE="exon",
            CODE="code",
            STRAND="strand",
        )

        lib = pd.read_csv("{}/{}".format(cls.DPATH, dfile), index_col=0).reset_index()

        lib = lib.rename(columns=r_cols)

        lib["chr"] = lib["chr"].apply(
            lambda v: f"chr{v}" if str(v) != "nan" else np.nan
        )

        return lib

    @classmethod
    def get_adam_core_essential(cls, dfile="gene_sets/pancan_core.csv"):
        return set(
            pd.read_csv(f"{cls.DPATH}/{dfile}")["ADAM PanCancer Core-Fitness genes"]
        )

    @classmethod
    def get_sanger_essential(cls, dfile="gene_sets/essential_sanger_depmap19.tsv"):
        return pd.read_csv(f"{cls.DPATH}/{dfile}", sep="\t")

    @classmethod
    def get_broad_core_essential(
        cls, dfile="gene_sets/pan_core_essential_broad_depmap18Q4.txt"
    ):
        return set(
            pd.read_csv("{}/{}".format(cls.DPATH, dfile))
            .iloc[:, 0]
            .apply(lambda v: v.split(" ")[0])
        )

    @classmethod
    def get_dummy_aa_vep(cls, dfile="bedit/aa_vep_jak1.csv"):
        return pd.read_csv("{}/{}".format(cls.DPATH, dfile))

    @classmethod
    def get_cytobands(cls, dfile="cytoBand.txt", chrm=None):
        cytobands = pd.read_csv(f"{cls.DPATH}/{dfile}", sep="\t")
        cytobands["Chr"] = [i.replace("chr", "") for i in cytobands["Chr"]]

        if chrm is not None:
            cytobands = cytobands[cytobands["Chr"] == chrm]

        assert cytobands.shape[0] > 0, f"{chrm} not found in cytobands file"

        return cytobands

    @classmethod
    def import_brass_bedpe(cls, bedpe_file, bkdist=None, splitreads=True):
        # Import BRASS bedpe
        bedpe_df = pd.read_csv(
            bedpe_file, sep="\t", names=cls.BRASS_HEADERS, comment="#"
        )

        # Correct sample name
        bedpe_df["sample"] = bedpe_df["sample"].apply(lambda v: v.split(",")[0])

        # SV larger than threshold
        if bkdist is not None:
            bedpe_df = bedpe_df[bedpe_df["bkdist"] >= bkdist]

        # BRASS2 annotated SV
        if splitreads:
            bedpe_df = bedpe_df.query("assembly_score != '_'")

        # Parse chromosome name
        bedpe_df = bedpe_df.assign(
            chr1=bedpe_df["chr1"].apply(lambda x: "chr{}".format(x)).values
        )
        bedpe_df = bedpe_df.assign(
            chr2=bedpe_df["chr2"].apply(lambda x: "chr{}".format(x)).values
        )

        return bedpe_df

    @staticmethod
    def gkn(values, bound=1e7):
        kernel = gaussian_kde(values)
        kernel = pd.Series(
            {
                k: np.log(
                    kernel.integrate_box_1d(-bound, v)
                    / kernel.integrate_box_1d(v, bound)
                )
                for k, v in values.to_dict().items()
            }
        )
        return kernel

    @staticmethod
    def two_vars_correlation(var1, var2, idx_set=None, method="pearson", verbose=0):
        if verbose > 0:
            print(f"Var1={var1.name}; Var2={var2.name}")

        if idx_set is None:
            idx_set = set(var1.dropna().index).intersection(var2.dropna().index)

        else:
            idx_set = set(var1.reindex(idx_set).dropna().index).intersection(
                var2.reindex(idx_set).dropna().index
            )

        if method == "spearman":
            r, p = spearmanr(
                var1.reindex(index=idx_set), var2.reindex(index=idx_set), nan_policy="omit"
            )
        else:
            r, p = pearsonr(var1.reindex(index=idx_set), var2.reindex(index=idx_set))

        return dict(corr=r, pval=p, len=len(idx_set))


class DotDict(dict):
    __getattr__ = dict.get
    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__



#!/usr/bin/env python
# Copyright (C) 2019 Emanuel Goncalves

import logging
import numpy as np
import pandas as pd
from scipy.stats import chi2
from sklearn.preprocessing import StandardScaler
from sklearn.linear_model import LinearRegression
from statsmodels.stats.multitest import multipletests


LOG = logging.getLogger("Crispy")


class LModel:
    def __init__(
        self,
        Y,
        X,
        M,
        M2=None,
        normalize=False,
        fit_intercept=True,
        copy_X=True,
        n_jobs=4,
        verbose=1,
    ):
        self.samples = set.intersection(
            set(Y.index),
            set(X.index),
            set(M.index),
            set(Y.index) if M2 is None else set(M2.index),
        )

        self.X = X.loc[self.samples]
        self.X = self.X.loc[:, self.X.count() > (M.shape[1] + (1 if M2 is None else 2))]
        self.X_ma = np.ma.masked_invalid(self.X.values)

        self.Y = Y.loc[self.samples]
        self.Y = self.Y.loc[:, self.Y.std() > 0]

        self.M = M.loc[self.samples]

        self.M2 = M2.loc[self.samples, self.X.columns] if M2 is not None else M2

        self.normalize = normalize
        self.fit_intercept = fit_intercept
        self.copy_X = copy_X
        self.n_jobs = n_jobs

        self.verbose = verbose
        self.log = logging.getLogger("Crispy")

    def model_regressor(self):
        regressor = LinearRegression(
            fit_intercept=self.fit_intercept,
            normalize=self.normalize,
            copy_X=self.copy_X,
            n_jobs=self.n_jobs,
        )
        return regressor

    @staticmethod
    def loglike(y_true, y_pred):
        nobs = len(y_true)
        nobs2 = nobs / 2.0

        ssr = np.power(y_true - y_pred, 2).sum()

        llf = -nobs2 * np.log(2 * np.pi) - nobs2 * np.log(ssr / nobs) - nobs2

        return llf

    @staticmethod
    def multipletests_per(
        associations, method="fdr_bh", field="pval", fdr_field="fdr", index_cols=None
    ):
        index_cols = ["y_id"] if index_cols is None else index_cols

        d_unique = {tuple(i) for i in associations[index_cols].values}

        df = associations.set_index(index_cols)

        df = pd.concat(
            [
                df.loc[i]
                .assign(fdr=multipletests(df.loc[i, field], method=method)[1])
                .rename(columns={"fdr": fdr_field})
                for i in d_unique
            ]
        ).reset_index()

        return df

    def fit_matrix(self):
        lms = []

        for x_idx, x_var in enumerate(self.X):
            if self.verbose > 0:
                self.log.info(f"LM={x_var} ({x_idx})")

            # Mask NaNs
            x_ma = np.ma.mask_rowcols(self.X_ma[:, [x_idx]], axis=0)

            # Build matrices
            x = self.X.iloc[~x_ma.mask.any(axis=1), [x_idx]]
            y = self.Y.iloc[~x_ma.mask.any(axis=1), :]

            # Covariate matrix (remove invariable features and add noise)
            m = self.M.iloc[~x_ma.mask.any(axis=1), :]
            if self.M2 is not None:
                m2 = self.M2.iloc[~x_ma.mask.any(axis=1), [x_idx]]
                m = pd.concat([m2, m], axis=1)
            m = m.loc[:, m.std() > 0]
            m += np.random.normal(0, 1e-4, m.shape)

            # Fit covariate model
            lm_small = self.model_regressor().fit(m, y)
            lm_small_ll = self.loglike(y, lm_small.predict(m))

            # Fit full model: covariates + feature
            lm_full_x = np.concatenate([m, x], axis=1)
            lm_full = self.model_regressor().fit(lm_full_x, y)
            lm_full_ll = self.loglike(y, lm_full.predict(lm_full_x))

            # Log-ratio test
            lr = 2 * (lm_full_ll - lm_small_ll)
            lr_pval = chi2(1).sf(lr)

            # Assemble + append results
            res = pd.DataFrame(
                dict(
                    y_id=y.columns,
                    x_id=x_var,
                    n=y.attrs["nan_mask"].loc[y.columns, x.index].sum(1) if "nan_mask" in y.attrs else len(x),
                    beta=lm_full.coef_[:, -1],
                    lr=lr.values,
                    covs=m.shape[1],
                    pval=lr_pval,
                    fdr=multipletests(lr_pval, method="fdr_bh")[1],
                )
            )

            lms.append(res)

        lms = pd.concat(lms, ignore_index=True).sort_values("pval")

        return lms

    @staticmethod
    def lm_residuals(y, x, fit_intercept=True, add_intercept=False):
        # Prepare input matrices
        ys = y.dropna()

        xs = x.loc[ys.index].dropna()
        xs = xs.loc[:, xs.std() > 0]

        ys = ys.loc[xs.index]

        if ys.shape[0] <= xs.shape[1]:
            return None

        # Linear regression models
        lm = LinearRegression(fit_intercept=fit_intercept).fit(xs, ys)

        # Calculate residuals
        residuals = ys - lm.predict(xs) - lm.intercept_

        # Add intercept
        if add_intercept:
            residuals += lm.intercept_

        return residuals


class LMModels:
    """"
    Class to perform the linear regression models
    """ ""

    RES_ORDER = [
        "y_id",
        "x_id",
        "beta",
        "beta_se",
        "pval",
        "fdr",
        "nsamples",
        "ncovariates",
    ]

    def __init__(
        self,
        y,
        x,
        k=None,
        m=None,
        m2=None,
        x_feature_type="all",
        m2_feature_type="same_y",
        add_intercept=True,
        lik="normal",
        transform_y="scale",
        transform_x="scale",
        transform_m2="scale",
        x_min_events=None,
        institute=True,
        verbose=1,
    ):
        # Misc
        self.verbose = verbose
        self.x_feature_type = x_feature_type
        self.m2_feature_type = m2_feature_type
        self.add_intercept = add_intercept

        # LIMIX parameters
        self.lik = lik

        # Preprocessing steps
        self.transform_y = transform_y
        self.transform_x = transform_x
        self.transform_m2 = transform_m2
        self.x_min_events = x_min_events

        # Build random effects and covariates matrices
        self.k = self.kinship(x) if k is None else k.copy()
        self.m = self.define_covariates(institute=institute) if m is None else m.copy()

        # Samples overlap
        self.samples = list(
            set.intersection(
                set(y.index),
                set(x.index),
                set(self.m.index),
                set(self.k.index),
                set(y.index) if m2 is None else set(m2.index),
            )
        )
        LOG.info(f"Samples: {len(self.samples)}")

        # Y matrix
        self.y, self.y_columns = self.__build_y(y.copy())

        # X matrix
        self.x, self.x_columns = self.__build_x(x.copy())

        # Covariates
        self.m = self.m.loc[self.samples, self.m.std() > 0]
        self.m, self.m_columns = self.m.values, np.array(list(self.m.columns))

        # Random effects matrix
        self.k = self.k.loc[self.samples, self.samples].values

        LOG.info(
            f"Y: {self.y.shape[1]}; X: {self.x.shape[1]}; M: {self.m.shape[1]}; K: {self.k.shape[1]}"
        )

        # Second covariates matrix
        if m2 is not None:
            self.m2, self.m2_columns = self.__build_m2(m2.copy())
            LOG.info(f"M2: {self.m2.shape[1]}")

        else:
            self.m2, self.m2_columns = None, None

    def __build_y(self, y):
        """
        Method to build the y matrix.

        :param y:
        :return:
        """
        y_ = self.transform_matrix(y.loc[self.samples], t_type=self.transform_y)
        return y_.values, np.array(list(y_.columns))

    def __build_m2(self, m2):
        """
        Method to build the m2 matrix.
        :param m2:
        :return:
        """
        m2_ = self.transform_matrix(m2.loc[self.samples], t_type=self.transform_m2)
        return m2_.values, np.array(list(m2_.columns))

    def __build_x(self, x):
        """
        Method to build the x matrix.
        :param x:
        :return:
        """
        x_ = x.loc[self.samples, x.std() > 0]

        if self.x_min_events is not None:
            x_ = x_.loc[:, x_.sum() >= self.x_min_events]
        else:
            x_ = self.transform_matrix(x_, t_type=self.transform_x)

        return x_.values, np.array(list(x_.columns))

    def __prepare_inputs__(self, y_var):
        # Define samples with NaNs
        y_idx = list(self.y_columns).index(y_var)
        y_nans_idx = np.isnan(self.y[:, y_idx])

        if self.verbose > 0:
            LOG.info(f"y_id: {y_var} ({y_idx}); N samples: {sum(1 - y_nans_idx)}")

        # Remove NaNs from y
        y_ = self.y[y_nans_idx == 0][:, [y_idx]]

        # Subset X
        x_ = self.x[y_nans_idx == 0]

        if self.x_feature_type == "drop_y":
            if y_var not in self.x_columns:
                LOG.warning(f"[x_feature_type=drop_y] Y feature {y_idx} not in X")

            x_ = x_[:, self.x_columns != y_var]
            x_vars = self.x_columns[self.x_columns != y_var]

        elif self.x_feature_type == "same_y":
            if y_var not in self.x_columns:
                LOG.error(f"[x_feature_type=same_y] Y feature {y_idx} not in X")

            x_ = x_[:, self.x_columns == y_var]
            x_vars = self.x_columns[self.x_columns == y_var]

        else:
            x_vars = self.x_columns[np.std(x_, axis=0) > 0]
            x_ = x_[:, np.std(x_, axis=0) > 0]

        # Subset m
        m_ = self.m[y_nans_idx == 0]
        m_ = m_[:, np.std(m_, axis=0) > 0]

        if (self.m2 is not None) and (self.m2_feature_type == "same_y"):
            m_ = np.append(
                m_, self.m2[y_nans_idx == 0][:, self.m2_columns == y_var], axis=1
            )

        if self.add_intercept:
            m_ = np.insert(m_, m_.shape[1], values=1, axis=1)

        # Subset random effects matrix
        k_ = self.k[:, y_nans_idx == 0][y_nans_idx == 0, :]

        return y_, y_nans_idx, x_, x_vars, m_, k_

    @staticmethod
    def log_likelihood(y_true, y_pred):
        n = len(y_true)
        ssr = np.power(y_true - y_pred, 2).sum()
        var = ssr / n

        l = np.longdouble(1 / (np.sqrt(2 * np.pi * var))) ** n * np.exp(
            -(np.power(y_true - y_pred, 2) / (2 * var)).sum()
        )
        ln_l = np.log(l)

        return float(ln_l)

    def lmm(self, y_var):
        """
        Linear regression method, using measurements of the y matrix for the variable specified by y_var.

        :param y_var: String y variable name
        :return: pandas.DataFrame of the associations
        """
        import limix

        y_, y_nans_idx, x_, x_vars, m_, k_ = self.__prepare_inputs__(y_var)

        # Linear Mixed Model
        lmm = limix.qtl.scan(G=x_, Y=y_, K=k_, M=m_, lik=self.lik, verbose=False)

        # Build results
        lmm_betas = lmm.effsizes["h2"].query("effect_type == 'candidate'")
        lmm = pd.DataFrame(
            dict(
                y_id=y_var,
                x_id=x_vars,
                beta=list(lmm_betas["effsize"].round(5)),
                beta_se=list(lmm_betas["effsize_se"].round(5)),
                pval=list(lmm.stats.loc[lmm_betas["test"], "pv20"]),
                nsamples=sum(1 - y_nans_idx),
                ncovariates=m_.shape[1],
            )
        )

        return lmm

    def matrix_lmm(self, pval_adj="fdr_bh", pval_adj_overall=False):
        # Iterate through Y variables
        res = []

        for y_var in self.y_columns:
            res.append(self.lmm(y_var=y_var))

        res = pd.concat(res, ignore_index=True)

        # Multiple p-value correction
        if pval_adj_overall:
            res = res.assign(fdr=multipletests(res["pval"], method=pval_adj)[1])

        else:
            res = self.multipletests(res, field="pval", pval_method=pval_adj)

        return res.sort_values("fdr")[self.RES_ORDER]

    def write_lmm(self, output_folder):
        for i in self.y_columns:
            self.lmm(y_var=i).to_csv(
                f"{output_folder}/{i}.csv.gz", index=False, compression="gzip"
            )

    @staticmethod
    def multipletests(
        parsed_results, pval_method="fdr_bh", field="pval", idx_cols=None
    ):
        idx_cols = ["y_id"] if idx_cols is None else idx_cols

        parsed_results_adj = []

        for idx, df in parsed_results.groupby(idx_cols):
            df = df.assign(fdr=multipletests(df[field], method=pval_method)[1])
            parsed_results_adj.append(df)

        parsed_results_adj = pd.concat(parsed_results_adj, ignore_index=True)

        return parsed_results_adj

    @staticmethod
    def transform_matrix(matrix, t_type="scale", add_nan_mask=True, fillna_func=np.mean):
        # Create mask attribute
        mask_df = matrix.notna()

        # Fill NaNs
        if fillna_func is not None:
            matrix = matrix.T.fillna(matrix.apply(fillna_func, axis=1)).T

        # Type of transformation
        if t_type == "scale":
            from sklearn.preprocessing import StandardScaler

            matrix = pd.DataFrame(
                StandardScaler().fit_transform(matrix),
                index=matrix.index,
                columns=matrix.columns,
            )

        elif t_type == "rank":
            matrix = matrix.rank(axis=1).values

        else:
            LOG.warning(
                f"{t_type} transformation not supported. Original matrix returned."
            )

        if add_nan_mask:
            matrix.attrs["nan_mask"] = mask_df.loc[matrix.index, matrix.columns]

        return matrix

    @staticmethod
    def define_covariates(
        std_filter=True,
        medium=True,
        tissuetype=True,
        cancertype=True,
        mburden=True,
        ploidy=True,
        institute=True,
    ):
        from crispy.DataImporter import Sample

        # Imports
        samplesheet = Sample().samplesheet

        # Covariates
        covariates = []

        # CRISPR institute of origin
        if type(institute) is pd.Series:
            covariates.append(pd.get_dummies(institute).astype(int))

        elif institute is True:
            covariates.append(pd.get_dummies(samplesheet["institute"]).astype(int))

        # Cell lines culture conditions
        if medium:
            culture = pd.get_dummies(samplesheet["growth_properties"]).drop(
                columns=["Unknown"]
            )
            covariates.append(culture)

        # Cancer type
        if cancertype:
            ctype = pd.get_dummies(samplesheet["cancer_type"])
            covariates.append(ctype)

        # Cancer type
        if tissuetype:
            ttype = pd.get_dummies(samplesheet["tissue"])
            covariates.append(ttype)

        # Mutation burden
        if mburden:
            m_burdern = samplesheet["mutational_burden"]
            covariates.append(m_burdern)

        # Ploidy
        if ploidy:
            ploidy = samplesheet["ploidy"]
            covariates.append(ploidy)

        # Merge covariates
        covariates = pd.concat(covariates, axis=1, sort=False)

        # Remove covariates with zero standard deviation
        if std_filter:
            covariates = covariates.loc[:, covariates.std() > 0]

        return covariates.dropna().astype(np.float)

    @staticmethod
    def kinship(k, decimal_places=5, kinship=False):
        if kinship:
            from limix.stats import linear_kinship

            K = pd.DataFrame(linear_kinship(k.values), index=k.index, columns=k.index)

        else:
            K = k.dot(k.T)
            K /= K.values.diagonal().mean()

        return K.round(decimal_places)



#!/usr/bin/env python
# Copyright (C) 2019 Emanuel Goncalves

import logging
import pandas as pd
import pkg_resources

# # sgRNAs library selection
#
#   Defining rules:
#       1) Protein-coding transcripts
#           + [Avana] CCDS: shortest transcript selected
#           + [Brunello] GENCODE: (i) APPRIS annotation; (ii) transcript confidence; (iii) support in GENCODE
#               + gene_type = "protein_coding"
#               + level = 1
#               + transcript_support_level = "1"
#               + tag appris_principal_ = "appris_principal_1"
#               (run pyensembl install --release 94 --species homo_sapiens)
#
#       2) Genes with multiple CCDS IDs, the shortest transcript was chosen
#
#       3) NGG protospacer adjacent motifs (PAMs) on both strands are annotated
#
#       4) sgRNA exclusion rules:
#           - Restriction sites
#               + [Avana + TKOv3] No BsmBI site: CGTCTC
#               + [TKOv3] No AgeI site: ACCGGT
#               + [TKOv3] No KpnI site: GGTACC
#               + [TKOv3] No BveI/BspMI site: ACCTGC
#               + [TKOv3] No BsmI site: GAATGC
#           - GC content
#               + [TKOv3] <= 40% or >= 75%
#           - Homopolymers
#               + [TKOv3] Any homopolymers length >= 4
#               + [Avana] T homopolymers length >= 4
#           - Germline SNP
#               + [TKOv3] No common SNP (db138) in PAM or sgRNA
#
#       5) Location in the protein coding sequence:
#           + [Avana] 0-25%: 57% of the guides; 25-50%: 43% of the guides
#
#       6) Off-target effect:
#           + [Avana] PAM-proximal (0) unique nts: -13 (84%); -17 (13%); -20 (0.2%); not unique (3%)
#           + [Yusa] sgRNAs with more than 1 perfect exonic match were excluded;
#           + [Yusa] N12NGG
#
#       7) Multi-guides:
#           + [Brunello] sgRNAs need to target a location at least 5% away from other guides from a protein-coding
#           standpoint.


class GuideDesign:
    BUILD = "Grch38"

    EXON_SEARCH = "https://www.sanger.ac.uk/htgt/wge/api/exon_search"
    CRISPR_SEARCH = "http://www.sanger.ac.uk/htgt/wge/api/crispr_search"
    CRISPR_BY_ID = "https://www.sanger.ac.uk/htgt/wge/api/crispr_by_id"
    SEARCH_BY_SEQ = "https://www.sanger.ac.uk/htgt/wge/api/search_by_seq"
    OFFTARGETS_BY_SEQ = "https://www.sanger.ac.uk/htgt/wge/api/off_targets_by_seq"
    INDIVIDUAL_OFF_TARGET_SEARCH = "https://www.sanger.ac.uk/htgt/wge/api/individual_off_target_search"

    RESTRICTION_SITES = dict(
        BsmBI="CGTCTC",
        AgeI="ACCGGT",
        KpnI="GGTACC",
        BveI="ACCTGC",
        BspMI="ACCTGC",
        BsmI="GAATGC",
    )

    def __init__(self, load_gencode=False):
        self.log = logging.getLogger("Crispy")

        self.dpath = pkg_resources.resource_filename("crispy", "data/")

        if load_gencode:
            # GENCODE: evidence based annotation of the human genome (GRCh38), version 29 (Ensembl 94)
            self.gencode = pd.read_csv(
                f"{self.dpath}/gencode.v29.annotation.gtf.gz",
                skiprows=5,
                sep="\t",
                header=None,
            )
            self.log.info(f"GENCODE annotation loaded: {self.gencode.shape}")

    @classmethod
    def individual_offtarget_search(cls, guide_id):
        offtargets = pd.read_json(
            f"{cls.INDIVIDUAL_OFF_TARGET_SEARCH}?ids[]={guide_id}&species={cls.BUILD}"
        )
        return offtargets

    @classmethod
    def offtargets_by_seq(cls, sequence, pam_right="false"):
        offtargets = pd.read_json(
            f"{cls.OFFTARGETS_BY_SEQ}?pam_right={pam_right}&species=human&seq={sequence}"
        )
        return offtargets

    @classmethod
    def search_by_seq(cls, seq, pam_right="2", get_db_data="1"):
        """

        :param seq:
        :param pam_right:
            2 - Default (search in both orientations);
            0 - only find crisprs on the global negative strand
            1 - only find crisprs on the global positive strand
        :param get_db_data:
            1 to return the crispr data for each ID found
        :return:
        """
        transcripts = pd.read_json(
            f"{cls.SEARCH_BY_SEQ}?seq={seq}&pam_right={pam_right}&get_db_data={get_db_data}&species={cls.BUILD}"
        )
        return transcripts

    @classmethod
    def search_by_id(cls, sgrna_id):
        transcripts = pd.read_json(
            f"{cls.CRISPR_BY_ID}?species={cls.BUILD}&id={sgrna_id}"
        )
        return transcripts

    @classmethod
    def get_wes_transcripts(cls, gene_symbol):
        query = f"{cls.EXON_SEARCH}?marker_symbol={gene_symbol}&species={cls.BUILD}"

        query = pd.read_json(query)

        return query

    @classmethod
    def get_wes_sgrnas(cls, exons):
        query = "&".join(map(lambda v: f"exon_id[]={v}", exons))
        query = f"{cls.CRISPR_SEARCH}?&species={cls.BUILD}&{query}"

        query = pd.read_json(query, typ="series")

        return query



#!/usr/bin/env python
# Copyright (C) 2019 Emanuel Goncalves

import os
import logging
import numpy as np
import pandas as pd
import pkg_resources
import scipy.stats as st
from pandas import DataFrame
from crispy.Utils import Utils


LOG = logging.getLogger("Crispy")

DATA_DIR = pkg_resources.resource_filename("crispy", "data/crispr_rawcounts/")
LIBS_DIR = pkg_resources.resource_filename("crispy", "data/crispr_libs/")
MANIFESTS_DIR = pkg_resources.resource_filename("crispy", "data/crispr_manifests/")

DATASETS = {
    "Yusa_v1": dict(
        name="Yusa v1",
        read_counts="Yusa_v1_Score_readcount.csv.gz",
        library="Yusa_v1.csv.gz",
        plasmids=["ERS717283.plasmid"],
        exclude_samples=set(
            pd.read_csv(f"{MANIFESTS_DIR}/project_score_all_qc_failed_samples.csv")["sample"]
        ),
    ),
    "Yusa_v1.1": dict(
        name="Yusa v1.1",
        read_counts="Yusa_v1.1_Score_readcount.csv.gz",
        library="Yusa_v1.1.csv.gz",
        plasmids=["CRISPR_C6596666.sample"],
        exclude_samples=set(
            pd.read_csv(
                f"{MANIFESTS_DIR}/project_score_exclude_samples.csv", header=None
            )[0]
        ),
    ),
    "GeCKOv2": dict(
        name="GeCKO v2",
        read_counts="GeCKO2_Achilles_v3.3.8_readcounts.csv.gz",
        library="GeCKO_v2.csv.gz",
        plasmids=["pDNA_pXPR003_120K_20140624"],
        exclude_guides=set(
            pd.read_csv(
                f"{MANIFESTS_DIR}/GeCKO2_Achilles_v3.3.8_dropped_guides.csv.gz"
            )["sgRNA"]
        ),
    ),
    "Avana_DepMap19Q2": dict(
        name="Avana DepMap19Q2",
        read_counts="Avana_DepMap19Q2_readcount.csv.gz",
        library="Avana_v1.csv.gz",
        plasmids=pd.read_csv(
            f"{MANIFESTS_DIR}/Avana_DepMap19Q2_sample_map.csv.gz", index_col="sample"
        )["controls"].apply(lambda v: v.split(";")).to_dict(),
        exclude_guides=set(
            pd.read_csv(f"{MANIFESTS_DIR}/Avana_DepMap19Q2_dropped_guides.csv.gz")[
                "guide"
            ]
        ),
    ),
    "Avana_DepMap19Q3": dict(
        name="Avana DepMap19Q3",
        read_counts="Avana_DepMap19Q3_readcount.csv.gz",
        library="Avana_v1.csv.gz",
        plasmids=pd.read_csv(
            f"{MANIFESTS_DIR}/Avana_DepMap19Q3_sample_map.csv.gz",
            index_col="replicate_ID",
        )["controls"]
        .apply(lambda v: v.split(";"))
        .to_dict(),
        exclude_guides=set(
            pd.read_csv(
                f"{MANIFESTS_DIR}/Avana_DepMap19Q3_dropped_guides.csv", index_col=0
            ).index
        ),
    ),
    "Avana_DepMap20Q1": dict(
        name="Avana DepMap20Q1",
        read_counts="Avana_DepMap20Q1_readcount.csv.gz",
        library="Avana_v1.csv.gz",
        plasmids=pd.read_csv(
            f"{MANIFESTS_DIR}/Avana_DepMap20Q1_sample_map.csv.gz",
            index_col="replicate_ID",
        )["controls"]
        .apply(lambda v: v.split(";"))
        .to_dict(),
        exclude_guides=set(
            pd.read_csv(
                f"{MANIFESTS_DIR}/Avana_DepMap20Q1_dropped_guides.csv", index_col=0
            ).index
        ),
    ),
    "Avana_DepMap20Q2": dict(
        name="Avana DepMap20Q2",
        read_counts="Avana_DepMap20Q2_readcount.csv.gz",
        library="Avana_v1.csv.gz",
        plasmids=pd.read_csv(
            f"{MANIFESTS_DIR}/Avana_DepMap20Q2_sample_map.csv.gz",
            index_col="replicate_ID",
        )["controls"]
        .apply(lambda v: v.split(";"))
        .to_dict(),
        exclude_guides=set(
            pd.read_csv(
                f"{MANIFESTS_DIR}/Avana_DepMap20Q2_dropped_guides.csv", index_col=0
            ).index
        ),
    ),
    "Sabatini_Lander_AML": dict(
        name="Sabatini Lander AML",
        read_counts="Sabatini_Lander_v2_AML_readcounts.csv.gz",
        library="Sabatini_Lander_v3.csv.gz",
        plasmids={
            "P31/FUJ-final": ["P31/FUJ-initial"],
            "NB4 (replicate A)-final": ["NB4-initial"],
            "NB4 (replicate B)-final": ["NB4-initial"],
            "OCI-AML2-final": ["OCI-AML2-initial"],
            "OCI-AML3-final": ["OCI-AML3-initial"],
            "SKM-1-final": ["SKM-1-initial"],
            "EOL-1-final": ["EOL-1-initial"],
            "HEL-final": ["HEL-initial"],
            "Molm-13-final": ["Molm-13-initial"],
            "MonoMac1-final": ["MonoMac1-initial"],
            "MV4;11-final": ["MV4;11-initial"],
            "PL-21-final": ["PL-21-initial"],
            "OCI-AML5-final": ["OCI-AML5-initial"],
        },
    ),
    "Brunello_A375": dict(
        name="Brunello A375",
        read_counts="Brunello_A375_readcount.csv.gz",
        library="Brunello_v1.csv.gz",
        plasmids={
            "orig_tracr_A375_RepA": ["orig_tracr_A375_pDNA"],
            "orig_tracr_A375_RepB": ["orig_tracr_A375_pDNA"],
            "mod_tracr_A375_RepA": ["mod_tracr_A375_pDNA"],
            "mod_tracr_A375_RepB": ["mod_tracr_A375_pDNA"],
            "mod_tracr_A375_RepC": ["mod_tracr_A375_pDNA"],
        },
    ),
    "HT29_Dabraf": dict(
        name="HT-29 Dabrafenib CRISPR",
        read_counts="Yusa_v1.1_HT29_dabraf.csv.gz",
        library="Yusa_v1.1.csv.gz",
        plasmids=["Plasmid_v1.1"],
        exclude_guides={
            "DHRSX_CCDS35195.1_ex1_X:2161152-2161175:+_3-1",
            "DHRSX_CCDS35195.1_ex6_Y:2368915-2368938:+_3-3",
            "DHRSX_CCDS35195.1_ex4_X:2326777-2326800:+_3-2",
            "sgPOLR2K_1",
        },
    ),
    "Organoids": dict(
        name="Organoids",
        read_counts="Yusa_v1.1_organoids.csv.gz",
        library="Yusa_v1.1.csv.gz",
        plasmids=["Plasmid_v1.1"],
        exclude_guides={
            "DHRSX_CCDS35195.1_ex1_X:2161152-2161175:+_3-1",
            "DHRSX_CCDS35195.1_ex6_Y:2368915-2368938:+_3-3",
            "DHRSX_CCDS35195.1_ex4_X:2326777-2326800:+_3-2",
            "sgPOLR2K_1",
        },
    ),
    "KM12_coverage": dict(
        name="KM12 coverage",
        read_counts="KM12_coverage.csv.gz",
        library="Yusa_v1.1.csv.gz",
        plasmids=pd.read_excel(
            f"{MANIFESTS_DIR}/KM12_coverage_samplesheet.xlsx", index_col="sample"
        )["plasmid"].apply(lambda v: v.split(";")).to_dict(),
        exclude_guides={
            "DHRSX_CCDS35195.1_ex1_X:2161152-2161175:+_3-1",
            "DHRSX_CCDS35195.1_ex6_Y:2368915-2368938:+_3-3",
            "DHRSX_CCDS35195.1_ex4_X:2326777-2326800:+_3-2",
            "sgPOLR2K_1",
        },
    ),
}


class Library:
    @staticmethod
    def load_library(lib_file, set_index=True, remove_dup=False, sep=","):
        lib_path = f"{LIBS_DIR}/{lib_file}"

        assert os.path.exists(lib_path), f"CRISPR library {lib_file} not supported"

        clib = pd.read_csv(lib_path, sep=sep)

        if remove_dup:
            nmaps = clib.groupby("sgRNA_ID")["Gene"].agg(list)
            clib = clib[clib["sgRNA_ID"].isin(set(nmaps[nmaps.apply(len) == 1].index))]

        if set_index:
            clib = clib.set_index("sgRNA_ID")

        return clib

    @staticmethod
    def load_library_sgrnas(verbose=0):
        lib_files = os.listdir(LIBS_DIR)
        lib_files = filter(lambda v: not v.startswith("."), lib_files)

        sgrnas = []

        for f in lib_files:
            if verbose > 0:
                LOG.info(f"Loading library {f}")

            flib = pd.read_csv(f"{LIBS_DIR}/{f}")["sgRNA"]
            flib = flib.value_counts().rename("count")
            flib = flib.reset_index().rename(columns={"index": "sgRNA"})
            flib = flib.assign(lib=f.replace(".csv.gz", ""))

            sgrnas.append(flib)

        sgrnas = pd.concat(sgrnas).reset_index(drop=True)

        sgrnas = sgrnas.assign(len=sgrnas["sgRNA"].apply(len))

        return sgrnas


class ReadCounts(DataFrame):
    PSEUDO_COUNT = 1

    def __init__(self, data=None, index=None, columns=None, dtype=None, copy=False):
        super().__init__(data, index, columns, dtype, copy)

    @property
    def _constructor(self):
        return ReadCounts

    def norm_rpm(self, scale=1e6):
        factors = self.sum() / scale
        return self.divide(factors)

    def norm_mean(self):
        factors = self.sum().mean() / self.sum()
        return self.divide(factors)

    def norm_gmean(self):
        sgrna_gmean = st.gmean(self, axis=1)
        factors = self.divide(sgrna_gmean, axis=0).median()
        return self.divide(factors)

    def foldchange(self, controls):
        if type(controls) == dict:
            fc = self.add(self.PSEUDO_COUNT)
            fc = pd.DataFrame(
                {
                    c: fc[c].divide(fc[controls[c]].mean(1), axis=0)
                    for c in controls
                    if c in fc
                }
            )
            fc = fc.apply(np.log2)

        else:
            fc = (
                self.add(self.PSEUDO_COUNT)
                .divide(self[controls].mean(1), axis=0)
                .drop(controls, axis=1)
                .apply(np.log2)
            )

        return fc

    def remove_low_counts(self, controls, counts_threshold=30):
        return self[self[controls].mean(1) >= counts_threshold]

    def scale(self, essential=None, non_essential=None, metric=np.median):
        if essential is None:
            essential = Utils.get_essential_genes(return_series=False)

        if non_essential is None:
            non_essential = Utils.get_non_essential_genes(return_series=False)

        assert (
            len(essential.intersection(self.index)) != 0
        ), "DataFrame has no index overlapping with essential list"

        assert (
            len(non_essential.intersection(self.index)) != 0
        ), "DataFrame has no index overlapping with non essential list"

        essential_metric = metric(self.reindex(essential).dropna(), axis=0)
        non_essential_metric = metric(self.reindex(non_essential).dropna(), axis=0)

        return self.subtract(non_essential_metric).divide(
            non_essential_metric - essential_metric
        )


class CRISPRDataSet:
    def __init__(self, dataset, ddir=None, exclude_samples=None, exclude_guides=None):
        # Load data-set dict
        if type(dataset) is dict:
            self.dataset_dict = dataset

        else:
            assert (
                dataset in DATASETS
            ), f"CRISPR data-set {dataset} not supported: {DATASETS}"

            self.dataset_dict = DATASETS[dataset]

        self.ddir = DATA_DIR if ddir is None else ddir

        # Build object arguments
        self.plasmids = self.dataset_dict["plasmids"]

        self.lib = Library.load_library(self.dataset_dict["library"])

        data = pd.read_csv(
            f"{self.ddir}/{self.dataset_dict['read_counts']}", index_col=0
        )

        # Drop excluded samples
        if exclude_samples is not None:
            data = data.drop(exclude_samples, axis=1, errors="ignore")
            LOG.info(f"#(samples)={len(exclude_samples)} excluded")

        elif "exclude_samples" in self.dataset_dict:
            data = data.drop(
                self.dataset_dict["exclude_samples"], axis=1, errors="ignore"
            )
            LOG.info(f"#(samples)={len(self.dataset_dict['exclude_samples'])} excluded")

        # Drop excluded guides
        if exclude_guides is not None:
            data = data.drop(exclude_guides, axis=0, errors="ignore")
            self.lib = self.lib.drop(exclude_guides, axis=0, errors="ignore")
            LOG.info(f"#(guides)={len(exclude_guides)} excluded")

        elif "exclude_guides" in self.dataset_dict:
            data = data.drop(
                self.dataset_dict["exclude_guides"], axis=0, errors="ignore"
            )
            self.lib = self.lib.drop(
                self.dataset_dict["exclude_guides"], axis=0, errors="ignore"
            )
            LOG.info(f"#(guides)={len(self.dataset_dict['exclude_guides'])} excluded")

        self.counts = ReadCounts(data=data)

        LOG.info(f"#(sgRNAs)={self.lib.shape[0]}")
        LOG.info(f"#(samples)={self.counts.shape[1]}")

    def get_plasmids_counts(self):
        return self.counts[self.plasmids]
